#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_alchemy.interfaces import IAlchemyEngineUtility
from pyams_form.interfaces.form import IWidgetsSuffixViewletsManager
from pyams_skin.interfaces.viewlet import IToolbarAddingMenu, ITableItemColumnActionsMenu
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import VIEW_SYSTEM_PERMISSION, MANAGE_SYSTEM_PERMISSION
from pyams_zmi.layer import IAdminLayer
from z3c.form.interfaces import IDataExtractedEvent, DISPLAY_MODE
from zope.component.interfaces import ISite

# import packages
from pyams_alchemy.engine import PersistentAlchemyEngineUtility, get_user_session
from pyams_form.form import AJAXAddForm, ajax_config
from pyams_form.schema import CloseButton
from pyams_pagelet.pagelet import pagelet_config
from pyams_skin.table import BaseTable
from pyams_skin.viewlet.toolbar import ToolbarMenuItem
from pyams_template.template import template_config
from pyams_utils.registry import query_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config, Viewlet
from pyams_zmi.form import AdminDialogEditForm, AdminDialogAddForm
from pyams_zmi.zmi.control_panel import UtilitiesTable
from pyramid.events import subscriber
from z3c.form import field, button
from z3c.table.column import GetAttrColumn
from zope.interface import Interface, Invalid
from zope.schema import Text

from pyams_alchemy import _


@viewlet_config(name='add-sqlalchemy-engine.menu', context=ISite, layer=IAdminLayer,
                view=UtilitiesTable, manager=IToolbarAddingMenu, permission=MANAGE_SYSTEM_PERMISSION)
class AlchemyEngineAddMenu(ToolbarMenuItem):
    """SQLAlchemy engine add menu"""

    label = _("Add SQLAlchemy engine...")
    label_css_class = 'fa fa-fw fa-database'
    url = 'add-sqlalchemy-engine.html'
    modal_target = True


@pagelet_config(name='add-sqlalchemy-engine.html', context=ISite, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
@ajax_config(name='add-sqlalchemy-engine.json', context=ISite, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION,
             base=AJAXAddForm)
class AlchemyEngineAddForm(AdminDialogAddForm):
    """SQLAlchemy engine add form"""

    title = _("Utilities")
    legend = _("Add SQLAlchemy engine")
    icon_css_class = 'fa fa-fw fa-database'

    fields = field.Fields(IAlchemyEngineUtility)
    edit_permission = None

    def create(self, data):
        return PersistentAlchemyEngineUtility()

    def add(self, object):
        manager = self.context.getSiteManager()
        name = (object.name or '').lower() or '__DEFAULT__'
        manager['sql::{0}'.format(name)] = object

    def nextURL(self):
        return absolute_url(self.context, self.request, 'utilities.html')


@subscriber(IDataExtractedEvent, form_selector=AlchemyEngineAddForm)
def handle_new_engine_data_extraction(event):
    """Handle new engine data extraction"""
    manager = event.form.context.getSiteManager()
    name = event.data['name'] or ''
    if 'sql::{0}'.format(name.lower() or '__DEFAULT__') in manager:
        event.form.widgets.errors += (Invalid(_("Specified engine name is already used!")), )
    engine = query_utility(IAlchemyEngineUtility, name=name)
    if engine is not None:
        event.form.widgets.errors += (Invalid(_("An SQLAlchemy engine is already registered with this name!")), )


@pagelet_config(name='properties.html', context=IAlchemyEngineUtility, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
@ajax_config(name='properties.json', context=IAlchemyEngineUtility, layer=IPyAMSLayer)
class AlchemyEnginePropertiesEditForm(AdminDialogEditForm):
    """SQLAlchemy engine properties edit form"""

    prefix = 'engine_properties.'

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("SQLAlchemy engine: {0}")).format(self.context.name)

    legend = _("Update SQLAlchemy engine properties")
    icon_css_class = 'fa fa-fw fa-database'

    fields = field.Fields(IAlchemyEngineUtility)
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(AlchemyEnginePropertiesEditForm, self).updateWidgets(prefix)
        self.widgets['name'].mode = DISPLAY_MODE
        self.widgets['encoding'].addClass('select2')


@viewlet_config(name='test-engine.menu', context=IAlchemyEngineUtility, layer=IAdminLayer,
                view=UtilitiesTable, manager=ITableItemColumnActionsMenu, permission=MANAGE_SYSTEM_PERMISSION)
class AlchemyEngineTestMenu(ToolbarMenuItem):
    """SQLAlchemy engine test menu"""

    label = _("Test connection...")
    label_css_class = 'fa fa-fw fa-play'
    url = 'test-sqlalchemy-engine.html'
    modal_target = True
    stop_propagation = True


class IAlchemyEngineTestFields(Interface):
    """SQLAlchemy engine test fields"""

    query = Text(title=_("SQL query"),
                 required=True)


class IAlchemyEngineTestButtons(Interface):
    """SQLAlchemy engine test form buttons"""

    close = CloseButton(name='close', title=_("Close"))
    execute = button.Button(name='execute', title=_("Execute SQL"))


@pagelet_config(name='test-sqlalchemy-engine.html', context=IAlchemyEngineUtility, layer=IPyAMSLayer,
                permission=MANAGE_SYSTEM_PERMISSION)
@ajax_config(name='test-sqlalchemy-engine.json', context=IAlchemyEngineUtility, layer=IPyAMSLayer, base=AJAXAddForm)
class AlchemyEngineTestForm(AdminDialogAddForm):
    """SQLAlchemy engine test form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("SQLAlchemy engine: {0}")).format(self.context.name)

    legend = _("Test SQLAlchemy engine")
    icon_css_class = 'fa fa-fw fa-database'

    fields = field.Fields(IAlchemyEngineTestFields)
    buttons = button.Buttons(IAlchemyEngineTestButtons)
    edit_permission = MANAGE_SYSTEM_PERMISSION

    @property
    def form_target(self):
        return '#{0}_test_result'.format(self.id)

    def updateActions(self):
        super(AlchemyEngineTestForm, self).updateActions()
        if 'execute' in self.actions:
            self.actions['execute'].addClass('btn-primary')

    def createAndAdd(self, data):
        data = data.get(self, data)
        session = get_user_session(self.context.name)
        return session.execute(data.get('query'))

    def get_ajax_output(self, changes):
        result = AlchemyEngineTestResults(self.context, self.request, changes)
        result.update()
        return {
            'status': 'success',
            'content': {
                'html': result.render()
            },
            'close_form': False
        }


@viewlet_config(name='engine-test-suffix', layer=IAdminLayer, manager=IWidgetsSuffixViewletsManager,
                view=AlchemyEngineTestForm, weight=50)
@template_config(template='templates/engine-test.pt')
class AlchemyEngineTestSuffix(Viewlet):
    """SQLAlchemy engine test suffix"""


class AlchemyEngineTestResults(BaseTable):
    """Alchemy engine test results table"""

    title = _("Query results")

    sortOn = None
    values = None

    def __init__(self, context, request, values):
        super(AlchemyEngineTestResults, self).__init__(context, request)
        self.values = list(values)

    @property
    def data_attributes(self):
        attrs = super(AlchemyEngineTestResults, self).data_attributes
        attrs['table'] = {
            'data-ams-datatable-global-filter': 'false',
            'data-ams-datatable-pagination-size': 'false'
        }
        return attrs

    def initColumns(self):
        if not self.values:
            self.columns = None
        else:
            columns = []
            row = self.values[0]
            for col in row.keys():
                column = GetAttrColumn(self.context, self.request, self)
                column.__name__ = col
                column.header = col
                column.attrName = col
                columns.append(column)
            self.columns = columns
